import torch
from torch import nn
from . import Model, manipulators


class ResConv(Model):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int = 3, stride: int = 1,
                 stimuli=nn.Hardswish):
        """
        conv block with res connection
        :param in_channels:
        :param out_channels:
        :param kernel_size:
        :param stride:
        """

        super().__init__()
        self.res_conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=1, stride=stride ** 2)
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride,
                      padding=kernel_size // 2),
            stimuli()
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(in_channels=out_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride,
                      padding=kernel_size // 2),
            stimuli()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        direct = self.conv2(self.conv1(x))
        res = self.res_conv(x)
        return res + direct


class ResTConv(Model):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int = 3, stride: int = 1,
                 stimuli=nn.Hardswish):
        """
        conv block with res connection
        :param in_channels:
        :param out_channels:
        :param kernel_size:
        :param stride:
        """

        super().__init__()
        self.res_conv = nn.ConvTranspose2d(in_channels=in_channels, out_channels=out_channels, kernel_size=1,
                                           stride=stride ** 2)
        self.conv1 = nn.Sequential(
            nn.ConvTranspose2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size,
                               stride=stride,
                               padding=kernel_size // 2),
            stimuli()
        )
        self.conv2 = nn.Sequential(
            nn.ConvTranspose2d(in_channels=out_channels, out_channels=out_channels, kernel_size=kernel_size,
                               stride=stride,
                               padding=kernel_size // 2),
            stimuli()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        direct = self.conv2(self.conv1(x))
        res = self.res_conv(x)
        return res + direct


class MultiplyLayer(manipulators.Multiply):
    def __init__(self, input_dim, output_dim, bias=True):
        route1 = nn.Linear(input_dim, output_dim, bias=bias)
        route2 = nn.Linear(input_dim, output_dim, bias=bias)
        super().__init__(route1, route2)



from torch import nn

from . import Model, blocks, manipulators


class GaussianModel(Model):
    def __init__(self):
        super().__init__(
            manipulators.SUM(
                Model(
                    blocks.MultiplyLayer(1, 1, False),
                    manipulators.Exp(),
                ),
            ),
            nn.Linear(1, 1, False)
        )



from torch import nn
import torch
import torch.utils
import torch.utils.data
from .torch_utils import persistence
import pickle

@persistence.persistent_class
class Model(nn.Sequential):
    """
    1. model with device param for device automatic adaptation.
    2. model with save and load method for model persistence.
    3. model with param_num property for model parameter count.
    4. model with output property for forward output.
    """
    
    def __init__(self, *layers):
        """
        init model with layers.
        """
        super().__init__()
        for idx, module in enumerate(layers):
            self.add_module(f'layer{idx}', module)
        
    def forward(self, x):
        """
        forward pass

        Args:
            x (Tensor): input tensor

        Returns:
            Tensor: output tensor
        """
        self.output = super().forward(x)
        return self.output

    #region device
    @ property
    def device(self):
        return next(self.parameters()).device
    #endregion

    #region save and load
    def save(self, path):
        with open(path.ensure(), 'wb') as f:
            pickle.dump(self, f)

    def load(self, path):
        return pickle.load(open(path, 'rb'))
    #endregion

    #region param count
    @ property
    def param_num(self):
        params = list(self.parameters())
        k = 0
        for i in params:
            l = 1
            for j in i.size():
                l *= j
            k = k + l
        return k

    #endregion

@persistence.persistent_class
class Dataset(torch.utils.data.Dataset):

    #region iterable
    def __iter__(self):
        self.iter_count = -1
        return self

    def __next__(self):
        self.iter_count += 1
        if self.iter_count >= len(self):
            raise StopIteration
        return self[self.iter_count]
    #endregion

    def get_loader(self, batch_size=4, shuffle=True):
        return torch.utils.data.DataLoader(self, batch_size=batch_size, shuffle=shuffle)


# -*- coding:utf-8 -*-
import sys
sys.argv.append('sdist')
from distutils.core import setup
from setuptools import find_packages

setup(name='EsyMod',
            version='39.2024.9.30.17.47',
            packages=['EsyMod',],
            description='a python lib for project files',
            long_description='',
            author='Quanfa',
            package_data={
            '': ['*.*'],
            },
            author_email='quanfa@tju.edu.cn',
            url='http://www.xxxxx.com/',
            license='MIT',
            )

            


import torch
from torch.nn import Flatten, Unflatten, Identity
from . import Model


class View(torch.nn.Module):
    def __init__(self, *shape):
        super().__init__()
        self.shape = shape

    def forward(self, x):
        id = 0
        shape = []
        for s in self.shape:
            if s == -1:
                shape.append(x.size(id))
                id += 1
            else:
                shape.append(s)
        return x.view(*shape)


#region math
class SUM(Model):
    def __init__(self, *routes):
        super().__init__(*routes)
        self.routes = routes

    def manipulate(self, x1, x2):
        return x1 + x2

    def forward(self, x):
        rs = [route(x) for route in self.routes]
        result = rs[0]
        for r in rs[1:]:
            result = self.manipulate(result, r)
        self.output = result
        return result


class Multiply(SUM):
    def manipulate(self, x1, x2):
        return x1 * x2


class MatrixMultiply(Multiply):
    def manipulate(self, x1, x2):
        return x1 @ x2


class Exp(torch.nn.Module):
    def forward(self, x):
        return torch.exp(x)
#endregion
